from __future__ import annotations

from collections import OrderedDict
from typing import Callable

import numpy as np
import PIL
import torch
from PIL.Image import Image
from torchvision import transforms

from qai_hub_models.models.fcn_resnet50.model import NUM_CLASSES
from qai_hub_models.utils.draw import create_color_map
from qai_hub_models.utils.image_processing import normalize_image_transform


def preprocess_image(image: Image) -> torch.Tensor:
    """
    Preprocesses images to be run through torch FCN segmenter
    as prescribed here:
    https://pytorch.org/hub/pytorch_vision_resnet/

    Parameters:
        image: Input image to be run through the classifier model.

    Returns:
        torch tensor to be directly passed to the model.
    """
    transform = transforms.Compose(
        [
            transforms.ToTensor(),
            normalize_image_transform(),
        ]
    )
    out_tensor: torch.Tensor = transform(image)  # type: ignore
    return out_tensor.unsqueeze(0)


class FCN_ResNet50App:
    """
    This class consists of light-weight "app code" that is required to
    perform end to end inference with FCN_ResNet50.

    For a given image input, the app will:
        * Pre-process the image (normalize)
        * Run image segmentation
        * Convert the raw output into probabilities using softmax
    """

    def __init__(self, model: Callable[[torch.Tensor], OrderedDict]):
        self.model = model

    def predict(self, image: Image, raw_output: bool = False) -> Image | np.ndarray:
        """
        From the provided image or tensor, segment the image

        Parameters:
            image: A PIL Image in RGB format.

        Returns:
            If raw_output is true, returns:
                masks: np.ndarray
                    A list of predicted masks.

            Otherwise, returns:
                segmented_images: List[PIL.Image]
                    Images with segmentation map overlaid with an alpha of 0.5.
        """

        input_tensor = preprocess_image(image)
        with torch.no_grad():
            output = self.model(input_tensor)
            output = output[0]
        predictions = output.argmax(0).byte().cpu().numpy()

        if raw_output:
            return predictions

        color_map = create_color_map(NUM_CLASSES)
        out = PIL.Image.blend(image, PIL.Image.fromarray(color_map[predictions]), 0.5)

        return out
